class SL_Point : Actor
{
	Default
	{
		+FLATSPRITE
		Scale 1.0;
	}
	
	States
	{
	Spawn:
		APLS B -1;
		Stop;
	}
}

// smartlight event handler
class smartlight_EventHandler : EventHandler
{
	PlayerInfo player;
	int counter;
	bool bOn;

	void smartlightOn()
	{
		player.mo.A_AttachLightDef("FLIGHT1", "flight1");
		player.mo.A_AttachLightDef("FLIGHT2", "flight2");
		bOn = true;
		counter = 35 * 3; // check # of seconds to stay on
	}
	void smartlightOff()
	{
		player.mo.A_RemoveLight("FLIGHT1");
		player.mo.A_RemoveLight("FLIGHT2");
		bOn = false;
		counter = 0;
	}

	double NormalizeAngle(double lAngle)
	{
		if (lAngle > 180 || lAngle < -180)
		{
			lAngle = (180 - abs(lAngle)) * (lAngle < 1 ? -1 : 1);
		}
		return lAngle;
	}

	// compare view angle to SL_Point
	bool isfacing(Actor mo)
	{
		double plnorm = Actor.Normalize180(player.mo.angle);
		double mAngle = VectorAngle(mo.pos.x - player.mo.pos.x, mo.pos.y - player.mo.pos.y);
		double dAngle = abs(NormalizeAngle(plnorm - mAngle));

		return (dAngle < 75); // check facing angle
	}

	int getBrightest(Actor ptr)
	{
		int SECLIGHT = 0;
		int MINLIGHT = 1;
		int FLOORLIGHT = 2;

		return clamp(max(ptr.args[SECLIGHT], ptr.args[MINLIGHT], ptr.args[FLOORLIGHT]), 0, 255);
	}

	// find SL_Point based on proximity, orientation, and darkness
	SL_Point findSmartLighter(vector3 v3)
	{
		ThinkerIterator SL_PointFinder;
		SL_PointFinder = ThinkerIterator.Create("SL_Point");
		SL_Point idpoint;
		SL_Point ptr = null;
		int nearest = 640; // check distance
		int darkest = 256;
	
		while (idpoint = SL_Point(SL_PointFinder.Next()))
		{
			double pl2id = player.mo.Distance2D(idpoint);
			int light = getBrightest(idpoint);

			if (pl2id < nearest && isfacing(idpoint) && light < darkest)
			{
				ptr = idpoint;
				darkest = light;
				nearest = pl2id;
			}
		}
		return ptr;
	}

	// return SL_Point spawned
	Actor spawnPoint(vector3 v3, Actor center)
	{
		Actor ptr = Actor.Spawn("SL_Point", v3);
		// remove if it can NOT be seen from the center (will eliminate some concave corners)
		if (!center.CheckSight(ptr))
		{
			ptr.A_Remove(AAPTR_DEFAULT, RMVF_EVERYTHING, "SL_Point");
			ptr = null;
		}
		return ptr;
	}

	// define how "convex" the corner should be
	bool sharpCorner(double vabs)
	{
//		int am_angle = CVar.FindCvar("am_angle").GetInt();

		int am_angle = 45;
		return ((vabs > 90 - am_angle && vabs < 90 + am_angle) || (vabs > 270 - am_angle && vabs < 270 + am_angle));
	}

	override void WorldTick()
	{
		int SECLIGHT = 0;
		int MINLIGHT = 1;
		int FLOORLIGHT = 2;

		if (counter == 0) //off or cooled off
		{
			SL_Point ptr = findSmartLighter(player.mo.pos);
			if (ptr)
			{
				int light = getBrightest(ptr);
				if (light < 129 && !bOn) // check light
				{
					smartlightOn();
				}
				if (light > 128 && bOn)
				{
					smartlightOff();
				}
			}
		}
		else
		{
			counter--;
		}
	}
	override void WorldLoaded(WorldEvent e)
	{
		int SECLIGHT = 0;
		int MINLIGHT = 1;
		int FLOORLIGHT = 2;

		int BLOCKING = 1;                 	// bit 0x0001
		int TWOSIDED = 2;                 	// bit 0x0004
		int DONTPEGTOP = 4;               // bit 0x0008
		int DONTPEGBOTTOM = 8;		// bit 0x0010
		int INNER = 0;

		for(int i = 0; i < Level.Sectors.Size(); i++)
		{
			SecPlane floorplane = Level.Sectors[i].FloorPlane;
			int floorZ = int(floorplane.ZAtPoint((Level.Sectors[i].CenterSpot.x, Level.Sectors[i].CenterSpot.y)));

			vector3 v3 = (Level.Sectors[i].CenterSpot.x, Level.Sectors[i].CenterSpot.y, floorZ);
			
			Actor center = Actor.Spawn("SL_Point", v3);

			for (int ii = 0; ii < Level.Sectors[i].Lines.Size(); ii++)
			{
				vector2 vA1 = Level.Sectors[i].Lines[ii].V1.P;
				vector2 vA2 = Level.Sectors[i].Lines[ii].V2.P;
				double vaii = VectorAngle(vA1.x - vA2.x, vA1.y - vA2.y);
				int AB = 0;
				
				int f = Level.Sectors[i].lines[ii].flags;
				int peg_top = f & (TWOSIDED + DONTPEGTOP);
				int peg_bottom = f & (TWOSIDED + DONTPEGBOTTOM);
				int blocking = f & (BLOCKING);

				Side sidedef = Level.Sectors[i].Lines[ii].Sidedef[INNER];
				textureid botid = sidedef.GetTexture(sidedef.Bottom); // depends on good design
				bool hide = !botid.isNull();

				if (blocking || hide)
				{
					for (int j = 0; j < Level.Sectors[i].Lines.Size(); j++)
					{
						if (j != ii)
						{
							vector2 vB1 = Level.Sectors[i].Lines[j].V1.P;
							vector2 vB2 = Level.Sectors[i].Lines[j].V2.P;
							double vaj = VectorAngle(vB1.x - vB2.x, vB1.y - vB2.y);
							double vabs = abs(abs(vaii) - abs(vaj));

							if (vA1 == vB2 && sharpCorner(vabs))
							{
								v3 = (vA1.x, vA1.y, floorZ);
								Actor ptr = spawnPoint(v3, center);
								if (ptr)
								{
									ptr.args[SECLIGHT] = Level.Sectors[i].lightlevel;
									ptr.args[MINLIGHT] = Level.Sectors[i].FindMinSurroundingLight(ptr.args[SECLIGHT]);
									ptr.args[FLOORLIGHT] = Level.Sectors[i].GetFloorLight();
								}
								AB++;
							}
							if (vA2 == vB1 && sharpCorner(vabs))
							{
								v3 = (vA2.x, vA2.y, floorZ);
								Actor ptr = spawnPoint(v3, center);
								if (ptr)
								{
									ptr.args[SECLIGHT] = Level.Sectors[i].lightlevel;
									ptr.args[MINLIGHT] = Level.Sectors[i].FindMinSurroundingLight(ptr.args[SECLIGHT]);
									ptr.args[FLOORLIGHT] = Level.Sectors[i].GetFloorLight();
								}
								AB++;
							}
						}
						if (AB == 2)
						{
							break;
						}
					}
				}
			}
			center.A_Remove(AAPTR_DEFAULT, RMVF_EVERYTHING, "SL_Point");
			center = null;
		}
	}

	override void PlayerRespawned(PlayerEvent e)
	{
		player = players[e.PlayerNumber];
		bOn = false;
		counter = 0;
	}
	override void PlayerSpawned(PlayerEvent e)
	{
		player = players[e.PlayerNumber];
		bOn = false;
		counter = 0;
	}
}
